<?php

namespace Drupal\instagram_media\Service;

use Drupal;
use Drupal\Component\Serialization\Json;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Database\Connection;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\File\FileSystemInterface;
use Drupal\Core\Logger\LoggerChannelFactoryInterface;
use Exception;
use GuzzleHttp\ClientInterface;
use GuzzleHttp\Exception\RequestException;
use Psr\Log\LoggerInterface;

class InstagramMediaService {

  private const API_URL = 'https://graph.facebook.com/v21.0/';

  /**
   * The database connection.
   *
   * @var \Drupal\Core\Database\Connection
   */
  protected Connection $database;

  protected ConfigFactoryInterface $configFactory;

  /**
   * Guzzle HTTP client.
   *
   * @var \GuzzleHttp\ClientInterface
   */
  protected ClientInterface $httpClient;

  protected EntityTypeManagerInterface $entityTypeManager;

  protected FileSystemInterface $fileSystem;

  /**
   * Logger channel.
   *
   * @var \Psr\Log\LoggerInterface
   */
  protected LoggerInterface $logger;

  /**
   * Construct a forecast API client.
   *
   * @param \GuzzleHttp\ClientInterface $http_client
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   Entity manager service.
   * @param \Drupal\Core\File\FileSystemInterface $file_system
   *   File system service.
   * @param \Drupal\Core\Logger\LoggerChannelFactoryInterface $logger_factory
   *   Logger factory service.
   * @param \Drupal\Core\Database\Connection $database
   *  The database connection.
   * @param \Drupal\Core\Config\ConfigFactoryInterface $config_factory
   */
  public function __construct(
    ClientInterface $http_client,
    EntityTypeManagerInterface $entity_type_manager,
    FileSystemInterface $file_system,
    LoggerChannelFactoryInterface $logger_factory,
    Connection $database,
    ConfigFactoryInterface $config_factory
  ) {
    $this->httpClient = $http_client;
    $this->entityTypeManager = $entity_type_manager;
    $this->fileSystem = $file_system;
    $this->logger = $logger_factory->get('instagram_media');
    $this->database = $database;
    $this->configFactory = $config_factory;
  }

  /**
   * Updates Instagram media data for blocks with configured Instagram accounts.
   *
   * This method retrieves Instagram media and related data for blocks
   * that have valid configurations, and inserts the fetched data into
   * the corresponding database tables.
   *
   * @return array|null
   *   Returns NULL if no media data is processed or an array with results
   *   in case further modifications are required (currently always NULL).
   * @throws \GuzzleHttp\Exception\GuzzleException
   */
  public function updateInstagramMediaData(): ?array {
    // Retrieve IDs of Instagram media blocks with active configurations.
    $ids = $this->getInstagramMediaBlockIds();

    if (empty($ids)) {
      // Return NULL if no blocks are configured.
      return NULL;
    }

    $posts_to_insert = [];
    $links_to_insert = [];

    foreach ($ids as $bid) {
      // Get the configuration for the block by its ID.
      $config = $this->getBlockConfig($bid);

      if (empty($config)) {
        // Skip if the block configuration is empty or invalid.
        continue;
      }

      // Extract Instagram Business ID from the block configuration.
      $instagramBusinessId = $this->getInstagramBusinessIdFromConfig($config);

      if (!$instagramBusinessId) {
        // Skip blocks without a valid Instagram Business ID.
        continue;
      }

      // Fetch Instagram media posts for the given Business ID and token.
      $posts = $this->fetchInstagramMedia(
        $instagramBusinessId,
        $config['token'],
        $config['limit']
      );
      $this->prepareMediaDirectory($config['block_id'], $config['media_path']);

      // Prepare data for insertion into the 'instagram_media_posts' table.
      foreach ($posts as $post) {
        $media_url = $config['media_path'] . $this->getFileName($post['media_url']);
        $media_uri = $this->fetchAndSaveLocalMedia($post['media_url'],
          $media_url);

        if (empty($media_uri)) {
          continue;
        }

        $posts_to_insert[] = [
          'block_id' => $config['block_id'],
          'media_type' => $post['media_type'],
          'media_url' => $media_uri,
          'caption' => $this->removeInvalidCharacters($post['caption'] ?? ''),
          'like_count' => $post['like_count'],
          'comments_count' => $post['comments_count'],
          'timestamp' => strtotime($post['timestamp']),
          'permalink' => $post['permalink'],
        ];
      }

      // Fetch and prepare custom Instagram links if specified in the config.
      if ($config['post_caption'] || $config['insights'] || $config['links']) {
        $links = $this->getInstagramLinks(
          $instagramBusinessId,
          $config['token']
        );
        $links['block_id'] = $config['block_id'];
        $links_to_insert[] = $links;
      }
    }

    // Insert media post data into the 'instagram_media_posts' table.
    $this->processDatabaseInsertion(
      $this->database,
      'instagram_media_posts',
      $posts_to_insert
    );

    // Insert custom links into the 'instagram_media_links' table.
    $this->processDatabaseInsertion(
      $this->database,
      'instagram_media_links',
      $links_to_insert
    );

    // Return NULL as no additional output is currently required.
    return NULL;
  }

  /**
   * Retrieves IDs of blocks using the 'instagram_media' plugin.
   *
   * This method queries the `block` entity type to find all blocks
   * configured with the 'instagram_media' plugin.
   *
   * @return array
   *   An array of block IDs that use the 'instagram_media' plugin.
   */
  private function getInstagramMediaBlockIds(): array {
    return Drupal::entityQuery('block')
      ->condition('plugin', 'instagram_media')
      ->execute();
  }

  /**
   * Retrieves the configuration settings for a specific block.
   *
   * This method fetches the block configuration based on the block ID.
   *
   * @param string $bid
   *   The block ID.
   *
   * @return array|null
   *   The block's configuration settings as an array, or NULL if not found.
   */
  private function getBlockConfig(string $bid): ?array {
    return $this->configFactory->get('block.block.' . $bid)->get('settings');
  }

  /**
   * Retrieves the Instagram Business ID from the block configuration.
   *
   * This method first retrieves the Facebook Page ID using the provided token,
   * and then uses it to fetch the associated Instagram Business ID.
   *
   * @param array $config
   *   The block configuration containing the required token.
   *
   * @return string|null
   *   The Instagram Business ID if found, or NULL if not available.
   *
   * @throws \GuzzleHttp\Exception\GuzzleException
   *   Throws an exception if the HTTP request fails.
   */
  private function getInstagramBusinessIdFromConfig(array $config): ?string {
    $facebookPageId = $this->getFacebookPageId($config['token']);
    return $this->getInstagramBusinessId($facebookPageId, $config['token']);
  }

  /**
   * Retrieves the Facebook Page ID using the provided access token.
   *
   * Sends an HTTP GET request to the Facebook Graph API to fetch
   * the page ID associated with the given token.
   *
   * @param string $token
   *   The access token for authenticating the API request.
   *
   * @return string|null
   *   The Facebook Page ID if successful, or NULL if not found or on failure.
   * @throws \GuzzleHttp\Exception\GuzzleException
   */
  public function getFacebookPageId(string $token): ?string {
    try {
      // Send a request to the Graph API to fetch page data.
      $response = $this->httpClient->request('GET',
        self::API_URL . 'me/accounts', [
          'query' => [
            'access_token' => $token,
            'fields' => 'id',
          ],
          'timeout' => 10,
        ]);

      // Decode the response data.
      $data = Json::decode($response->getBody()->getContents());
      return $data['data'][0]['id'] ?? NULL;
    }
    catch (Exception $e) {
      // Log an error message if the request fails.
      $this->logger->error('Failed to retrieve Facebook Page ID: @message', [
        '@message' => $e->getMessage(),
      ]);
      return NULL;
    }
  }

  /**
   * Fetches the Instagram Business ID associated with a Facebook Page ID.
   *
   * This method sends a request to the Facebook Graph API to retrieve
   * the Instagram Business ID linked to the provided Facebook Page ID.
   *
   * @param string $facebookPageId
   *   The Facebook Page ID.
   * @param string $token
   *   The access token for the API request.
   *
   * @return string|null
   *   The Instagram Business ID if found, or NULL if not available.
   *
   * @throws \GuzzleHttp\Exception\GuzzleException
   *   Throws an exception if the HTTP request fails.
   */
  public function getInstagramBusinessId(
    string $facebookPageId,
    string $token
  ): ?string {
    try {
      // Send a request to the Graph API to fetch the Instagram Business Account.
      $response = $this->httpClient->request('GET',
        self::API_URL . $facebookPageId, [
          'query' => [
            'fields' => 'instagram_business_account',
            'access_token' => $token,
          ],
          'timeout' => 10,
        ]);

      // Decode the response and return the Instagram Business ID.
      $data = Json::decode($response->getBody()->getContents());
      return $data['instagram_business_account']['id'] ?? NULL;
    }
    catch (Exception $e) {
      // Log an error message if the request fails.
      $this->logger->error('Error retrieving Instagram Business ID: @message', [
        '@message' => $e->getMessage(),
      ]);
      return NULL;
    }
  }

  /**
   * Fetches Instagram media data for a given Instagram Business ID.
   *
   * This method retrieves media items such as posts, images, and videos
   * associated with the specified Instagram Business ID.
   *
   * @param string $instagramBusinessId
   *   The Instagram Business ID.
   * @param string $token
   *   The access token for the API request.
   * @param int $limit
   *   The number of media items to fetch. Defaults to 6, with a maximum of 25.
   *
   * @return array
   *   An array of media items, including details like media type, URL,
   *   caption, like count, comments count, and timestamp.
   *
   * @throws \GuzzleHttp\Exception\GuzzleException
   *   Throws an exception if the HTTP request fails.
   */
  public function fetchInstagramMedia(
    string $instagramBusinessId,
    string $token,
    int $limit = 6
  ): array {
    try {
      $limit = min($limit, 25);
      // Send a request to the Graph API to fetch media data.
      $response = $this->httpClient->request('GET',
        self::API_URL . "$instagramBusinessId/media", [
          'query' => [
            'fields' => 'id,media_type,media_url,permalink,caption,like_count,comments_count,timestamp,thumbnail_url',
            'access_token' => $token,
            'limit' => $limit,
            // Limit capped at 25 by API restrictions.
          ],
          'timeout' => 10,
        ]);

      // Decode the response and return the media data.
      $data = Json::decode($response->getBody())['data'] ?? [];

      return array_values(array_filter($data,
        static fn(array $item): bool => isset($item['media_url'])));
    }
    catch (Exception $e) {
      // Log an error message if the request fails.
      $this->logger->error('Error fetching Instagram media data: @message', [
        '@message' => $e->getMessage(),
      ]);
      return [];
    }
  }

  protected function removeInvalidCharacters(string $text): string {
    return mb_substr(preg_replace('/[^\x{0000}-\x{FFFF}]/u', '', $text), 0,
      255);
  }

  /**
   * Fetches Instagram profile links and data.
   *
   * This method retrieves profile information for the specified Instagram
   * Business ID, including username, profile picture, and follower statistics.
   *
   * @param string $instagramBusinessId
   *   The Instagram Business ID.
   * @param string $token
   *   The access token for the API request.
   *
   * @return array|null
   *   An array of Instagram profile data, including name, username, profile
   *   picture URL, media count, and follower/following stats. Returns NULL if
   *   an error occurs.
   * @throws \GuzzleHttp\Exception\GuzzleException
   */
  public function getInstagramLinks(
    string $instagramBusinessId,
    string $token
  ): ?array {
    try {
      // Send a request to the Graph API to fetch profile data.
      $response = $this->httpClient->request('GET',
        self::API_URL . $instagramBusinessId, [
          'query' => [
            'access_token' => $token,
            'fields' => 'name,username,profile_picture_url,media_count,followers_count,follows_count',
          ],
          'timeout' => 10,
        ]);

      // Decode the API response.
      $data = Json::decode($response->getBody()->getContents());

      // Provide a placeholder avatar if no profile picture is available.
      $profile_picture_url = $data['profile_picture_url'] ?? Drupal::service('file_url_generator')
        ->generateAbsoluteString(Drupal::service('extension.list.module')
            ->getPath('instagram_media') . '/misc/images/placeholder-avatar.jpg');

      // Build and return profile data.
      return [
        'name' => $this->removeInvalidCharacters($data['name'] ?? ''),
        'username' => $data['username'] ?? '',
        'profile_url' => $data['username'] ? 'https://www.instagram.com/' . $data['username'] . '/' : '',
        'profile_picture_url' => $profile_picture_url,
        'media_count' => $data['media_count'] ?? 0,
        'followers_count' => self::formatCount($data['followers_count'] ?? 0),
        'follows_count' => self::formatCount($data['follows_count'] ?? 0),
      ];
    }
    catch (Exception $e) {
      // Log an error message if the request fails.
      $this->logger->error('Error fetching Instagram profile data: @message', [
        '@message' => $e->getMessage(),
      ]);
      return NULL;
    }
  }

  /**
   * Formats large count values into human-readable strings.
   *
   * This method converts large numeric counts into more compact formats,
   * such as "1k" for 1,000 and "1M" for 1,000,000. This is useful for
   * displaying Instagram follower counts, media counts, etc., in a
   * user-friendly way.
   *
   * @param int $count
   *   The count value to be formatted (e.g., a number representing followers).
   *
   * @return string
   *   The formatted count as a string (e.g., "1k", "1.5M", "500").
   */
  private static function formatCount(int $count): string {
    // If count is greater than or equal to 1,000,000, format as "M" (million).
    if ($count >= 1000000) {
      return round($count / 1000000, 1) . 'M';
    }

    // If count is greater than or equal to 1,000, format as "k" (thousand).
    if ($count >= 1000) {
      return round($count / 1000, 1) . 'k';
    }

    // Otherwise, return the count as-is (for values less than 1,000).
    return (string) $count;
  }

  /**
   * Inserts data into a database table with transaction processing.
   *
   * This method ensures that data is inserted into the specified table,
   * and handles database operations within a transaction to maintain data
   * integrity. If an error occurs during the insertion process, the
   * transaction is rolled back.
   *
   * @param \Drupal\Core\Database\Connection $connection
   *   The database connection object to execute queries.
   * @param string $table
   *   The name of the table where the data will be inserted.
   * @param array $data
   *   An array of data rows to insert into the table. Each row is an
   *   associative array where the keys are column names, and the values are
   *   the corresponding values to insert.
   *
   * @throws \Exception
   *   If there is an error during the insertion process, an exception will be
   *   thrown and logged.
   */
  protected function processDatabaseInsertion(
    Connection $connection,
    string $table,
    array $data,
    $blockId = NULL
  ): void {
    // If no data is provided, exit early to avoid unnecessary database operations.
    if (empty($data)) {
      return;
    }

    // Start a new database transaction to ensure atomicity of the insertion process.
    $transaction = $connection->startTransaction();

    try {
      // If a block_id is provided, truncate only the rows matching the block_id.
      if ($blockId) {
        $connection->delete($table)
          ->condition('block_id', $blockId)
          ->execute();
      }
      else {
        // Otherwise, truncate the entire table.
        $connection->truncate($table)->execute();
      }

      // Insert each row of data into the specified table.
      foreach ($data as $row) {
        $connection->insert($table)
          ->fields($row) // Map the fields to insert based on the provided row data.
          ->execute(); // Execute the insertion.
      }
    }
    catch (Exception $e) {
      // If an error occurs, rollback the transaction to maintain data consistency.
      $transaction->rollback();

      // Log the error for debugging purposes.
      $this->logger->error('Failed to insert data into @table: @message', [
        '@table' => $table,
        '@message' => $e->getMessage(),
      ]);
    }

    // Unset the transaction object to free resources.
    unset($transaction);
  }

  /**
   * Submits Instagram media data and custom links into the database.
   *
   * This method fetches Instagram media posts and associated custom links,
   * then inserts the media posts and links into the respective database
   * tables.
   * It performs data retrieval, processing, and insertion for the Instagram
   * media block.
   *
   * @param string $token
   *   The access token for the Instagram account.
   * @param int $limit
   *   The number of media posts to fetch from Instagram.
   * @param string $blockId
   *   The ID of the block where the media will be displayed.
   * @param string|null $post_caption
   *   The caption for the post (optional).
   * @param mixed $insights
   *   The insights data (optional).
   * @param mixed $links
   *   The custom links (optional).
   *
   * @return array|null
   *   Returns NULL after the data is successfully processed and inserted.
   * @throws \GuzzleHttp\Exception\GuzzleException
   * @throws \Exception
   */
  public function submitInstagramMediaData(
    $token,
    $limit,
    $blockId,
    $post_caption,
    $insights,
    $links,
    $media_path
  ): ?array {
    // Initialize empty arrays for posts and links to be inserted into the database.
    $posts_to_insert = [];
    $links_to_insert = [];

    // Retrieve the Facebook Page ID and Instagram Business ID using the access token.
    $facebookPageId = $this->getFacebookPageId($token);
    $instagramBusinessId = $this->getInstagramBusinessId($facebookPageId,
      $token);

    // If no Instagram Business ID is found, return NULL.
    if (!$instagramBusinessId) {
      return NULL;
    }

    // Fetch Instagram media posts based on the Business ID and token.
    $posts = $this->fetchInstagramMedia($instagramBusinessId, $token, $limit);

    $this->prepareMediaDirectory($blockId, $media_path);

    // Process each fetched post and prepare data for insertion into the database.
    foreach ($posts as $post) {
      $media_url = $media_path . $this->getFileName($post['media_url']);
      $media_uri = $this->fetchAndSaveLocalMedia($post['media_url'],
        $media_url);

      if (empty($media_uri)) {
        continue;
      }

      $posts_to_insert[] = [
        'block_id' => $blockId,
        'media_type' => $post['media_type'],
        'media_url' => $media_uri,
        'caption' => $this->removeInvalidCharacters($post['caption'] ?? ''),
        'like_count' => $post['like_count'],
        'comments_count' => $post['comments_count'],
        'timestamp' => strtotime($post['timestamp']),
        'permalink' => $post['permalink'],
      ];
    }

    // If custom links or insights are provided, fetch Instagram profile links and prepare them for insertion.
    if ($post_caption || $insights || $links) {
      $links = $this->getInstagramLinks($instagramBusinessId, $token);
      $links['block_id'] = $blockId; // Associate the links with the current block ID.
      $links_to_insert[] = $links;
    }

    // Insert the processed Instagram media posts and links into their respective database tables.
    $this->processDatabaseInsertion($this->database, 'instagram_media_posts',
      $posts_to_insert, $blockId);
    $this->processDatabaseInsertion($this->database, 'instagram_media_links',
      $links_to_insert, $blockId);

    // Return NULL after successfully processing and inserting the data.
    return NULL;
  }

  /**
   * Prepare the media directory by clearing and recreating it.
   */
  protected function prepareMediaDirectory($blockId, $media_path): void {
    if (is_dir($media_path)) {
      $this->fileSystem->deleteRecursive($media_path);
      Drupal::logger('instagram_media')
        ->notice('Instagram Media media data for @block_id has been purged.',
          ['@block_id' => $blockId]);
    }
    $this->fileSystem->mkdir($media_path, NULL, TRUE);
  }

  /**
   * Get file name from URL.
   */
  protected function getFileName($url) {
    $info = pathinfo(parse_url($url)['path']);
    return $info['basename'];
  }

  public function fetchAndSaveLocalMedia(
    string $sourceUrl,
    string $destinationPath
  ): ?string {
    try {
      $fileContent = (string) \Drupal::httpClient()
        ->get($sourceUrl, ['timeout' => 10])
        ->getBody();

      if (empty($fileContent)) {
        Drupal::logger('instagram_media')
          ->warning('Empty content fetched from URL: @url',
            ['@url' => $sourceUrl]);
        return NULL;
      }

      $directory = dirname($destinationPath);
      if (!\Drupal::service('file_system')->prepareDirectory($directory,
        FileSystemInterface::CREATE_DIRECTORY)) {
        Drupal::logger('instagram_media')
          ->error('Failed to prepare directory: @directory',
            ['@directory' => $directory]);
        return NULL;
      }
      $file = \Drupal::service('file.repository')->writeData(
        $fileContent,
        $destinationPath,
        FileSystemInterface::EXISTS_REPLACE
      );

      if (!$file) {
        Drupal::logger('instagram_media')
          ->error('Failed to save file at path: @path',
            ['@path' => $destinationPath]);
        return NULL;
      }

      return $file->getFileUri();
    }
    catch (RequestException $e) {
      Drupal::logger('instagram_media')
        ->error('HTTP error occurred while fetching file from URL: @url. Error: @error',
          [
            '@url' => $sourceUrl,
            '@message' => $e->getMessage(),
          ]);
    }
    catch (\Exception $e) {
      Drupal::logger('instagram_media')
        ->error('Unexpected error occurred: @error',
          ['@error' => $e->getMessage()]);
      return NULL;
    }
    return NULL;
  }

}
